@@ -2,7 +2,8 @@
  * filterdiff - extract (or exclude) a diff from a diff file
  * lsdiff - show which files are modified by a patch
  * grepdiff - show files modified by a patch containing a regexp
- * Copyright (C) 2001, 2002, 2003, 2004, 2008 Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2001, 2002, 2003, 2004, 2008, 2009 Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2001, 2002, 2003, 2004, 2009 Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -83,6 +84,8 @@ static int annotating = 0;
 static int ignore_components = 0;
 static int strip_components = 0;
 static const char *prefix_to_add = NULL;
+static const char *old_prefix_to_add = NULL;
+static const char *new_prefix_to_add = NULL;
 static int show_status = 0;
 static int verbose = 0;
 static int removing_timestamp = 0;
@@ -156,6 +159,12 @@ static int output_header_line (const char *line)
 
 	if (prefix_to_add)
 		fputs (prefix_to_add, stdout);
+	else {
+		if (old_prefix_to_add && strncmp (line, "---", 3) == 0)
+			fputs (old_prefix_to_add, stdout);
+		if (new_prefix_to_add && strncmp (line, "+++", 3) == 0)
+			fputs (new_prefix_to_add, stdout);
+	}
 
 	fn = xstrndup (line + 4, h);
 	fputs (stripped (fn, strip_components), stdout);
@@ -284,7 +293,7 @@ do_unified (FILE *f, char *header[2], int match, char **line,
 	int orig_is_empty = 1, new_is_empty = 1; /* assume until otherwise */
 
 	if (output_matching == output_file)
-		match_tmpf = tmpfile ();
+		match_tmpf = xtmpfile ();
 
 	for (;;) {
 		ssize_t got = getline (line, linelen, f);
@@ -315,7 +324,7 @@ do_unified (FILE *f, char *header[2], int match, char **line,
 			if (output_matching == output_hunk) {
 				if (match_tmpf)
 					fclose (match_tmpf);
-				match_tmpf = tmpfile ();
+				match_tmpf = xtmpfile ();
 			}
 
 			if (read_atatline (*line, &orig_offset, &orig_count,
@@ -336,7 +345,7 @@ do_unified (FILE *f, char *header[2], int match, char **line,
 			else hunk_match = 0;
 
 			trailing = strchr (*line, '+');
-			trailing += strcspn (trailing, " ");
+			trailing += strcspn (trailing, " \n");
 			if (*trailing == ' ')
 				trailing++;
 			trailing += strspn (trailing, "@");
@@ -559,7 +568,7 @@ do_context (FILE *f, char *header[2], int match, char **line,
 		return 1;
 
 	if (output_matching == output_file)
-		match_tmpf = tmpfile ();
+		match_tmpf = xtmpfile ();
 
  next_hunk:
 	unchanged = 0;
@@ -599,7 +608,7 @@ do_context (FILE *f, char *header[2], int match, char **line,
 			if (output_matching == output_hunk) {
 				if (match_tmpf)
 					fclose (match_tmpf);
-				match_tmpf = tmpfile ();
+				match_tmpf = xtmpfile ();
 			}
 		}
 
@@ -1015,10 +1024,14 @@ static int filterdiff (FILE *f, const char *patchname)
 
 const char * syntax_str =
 "Options:\n"
-"  -x PAT    exclude files matching PAT\n"
-"  -X FILE   exclude files that match any pattern in FILE\n"
-"  -i PAT    include only files matching PAT\n"
-"  -I FILE   include only files that match any pattern in FILE\n"
+"  -x PAT, --exclude=PAT\n"
+"            exclude files matching PAT\n"
+"  -X FILE, --exclude-from-file=FILE\n"
+"            exclude files that match any pattern in FILE\n"
+"  -i PAT, --include=PAT\n"
+"            include only files matching PAT\n"
+"  -I FILE, --include-from-file=FILE\n"
+"            include only files that match any pattern in FILE\n"
 "  --hunks=H, -# H\n"
 "            include only hunks in range H\n"
 "  --lines=L include only hunks with (original) lines in range L\n"
@@ -1036,24 +1049,35 @@ const char * syntax_str =
 "            don't show timestamps from output (filterdiff, grepdiff)\n"
 "  --clean (filterdiff)\n"
 "            remove all comments (non-diff lines) from output (filterdiff)\n"
-"  -z        decompress .gz and .bz2 files\n"
-"  -n        show line numbers (lsdiff, grepdiff)\n"
+"  -z, --decompress\n"
+"            decompress .gz and .bz2 files\n"
+"  -n, --line-number\n"
+"            show line numbers (lsdiff, grepdiff)\n"
 "  --number-files (lsdiff, grepdiff)\n"
 "            show file numbers, for use with filterdiff's --files option (lsdiff, grepdiff)\n"
 "  -H, --with-filename (lsdiff, grepdiff)\n"
 "            show patch file names (lsdiff, grepdiff)\n"
 "  -h, --no-filename (lsdiff, grepdiff)\n"
 "            suppress patch file names (lsdiff, grepdiff)\n"
-"  -p N      initial pathname components to ignore\n"
+"  -p N, --strip-match=N\n"
+"            initial pathname components to ignore\n"
 "  --strip=N initial pathname components to strip\n"
 "  --addprefix=PREFIX\n"
 "            prefix pathnames with PREFIX\n"
-"  -s        show file additions and removals (lsdiff)\n"
-"  -v        verbose output -- use more than once for extra verbosity\n"
-"  -E        use extended regexps, like egrep (grepdiff)\n"
+"  --addoldprefix=PREFIX\n"
+"            prefix pathnames in old files with PREFIX\n"
+"  --addnewprefix=PREFIX\n"
+"            prefix pathnames in new files with PREFIX\n"
+"  -s, --status\n"
+"            show file additions and removals (lsdiff)\n"
+"  -v, --verbose\n"
+"            verbose output -- use more than once for extra verbosity\n"
+"  -E, --extended-regexp\n"
+"            use extended regexps, like egrep (grepdiff)\n"
 "  -E, --empty-files-as-absent (lsdiff)\n"
 "            treat empty files as absent (lsdiff)\n"
-"  -f FILE   read regular expressions from FILE (grepdiff)\n"
+"  -f FILE, --file=FILE\n"
+"            read regular expressions from FILE (grepdiff)\n"
 "  --filter  run as 'filterdiff' (grepdiff, lsdiff)\n"
 "  --list    run as 'lsdiff' (filterdiff, grepdiff)\n"
 "  --grep    run as 'grepdiff' (filterdiff, lsdiff)\n"
@@ -1250,6 +1274,8 @@ int main (int argc, char *argv[])
 			{"grep", 0, 0, 'g'},
 			{"strip", 1, 0, 1000 + 'S'},
 			{"addprefix", 1, 0, 1000 + 'A'},
+			{"addoldprefix", 1, 0, 1000 + 'O'},
+			{"addnewprefix", 1, 0, 1000 + 'N'},
 			{"hunks", 1, 0, '#'},
 			{"lines", 1, 0, 1000 + ':'},
 			{"files", 1, 0, 1000 + 'w'},
@@ -1263,6 +1289,18 @@ int main (int argc, char *argv[])
 			{"empty-files-as-absent", 0, 0, 'E'},
 			{"number-files", 0, 0, 1000 + 'n'},
 			{"clean", 0, 0, 1000 + 'c'},
+			{"strip-match", 1, 0, 'p'},
+			{"include", 1, 0, 'i'},
+			{"exclude", 1, 0, 'x'},
+			{"include-from-file", 1, 0, 'I'},
+			{"exclude-from-file", 1, 0, 'X'},
+			{"decompress", 0, 0, 'z'},
+			{"line-number", 0, 0, 'n'},
+			{"strip-match", 1, 0, 'p'},
+			{"status", 0, 0, 's'},
+			{"extended-regexp", 0, 0, 'E'},
+			{"empty-files-as-removed", 0, 0, 'E'},
+			{"file", 1, 0, 'f'},
 			{0, 0, 0, 0}
 		};
 		char *end;
@@ -1309,6 +1347,12 @@ int main (int argc, char *argv[])
 		case 1000 + 'A':
 			prefix_to_add = optarg;
 			break;
+		case 1000 + 'O':
+			old_prefix_to_add = optarg;
+			break;
+		case 1000 + 'N':
+			new_prefix_to_add = optarg;
+			break;
 		case 'p':
 			ignore_components = strtoul (optarg, &end, 0);
 			if (optarg == end)
