@@ -1,6 +1,6 @@
 /*
  * rediff - fix offset and counts of a hand-edited diff
- * Copyright (C) 2001, 2002 Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2001, 2002, 2004, 2009 Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -127,6 +127,7 @@ static unsigned long adjust_offsets_and_copy (long *offset, FILE *in,
 	size_t linelen = 0;
 	unsigned long orig_offset, orig_lines, new_offset, new_lines;
 	unsigned long count = 0;
+	char *trailing;
 
 	if (getline (&line, &linelen, in) == -1)
 		goto out;
@@ -154,7 +155,12 @@ static unsigned long adjust_offsets_and_copy (long *offset, FILE *in,
 		error (EXIT_FAILURE, 0, "Line not understood: %s", line);
 	}
 
-	free (line);
+	/* Find the part after "@@...@@". */
+	trailing = strchr (line, '+');
+	trailing += strcspn (trailing, " \n");
+	if (*trailing == ' ')
+		trailing++;
+	trailing += strspn (trailing, "@");
 
 	/* Adjust offsets */
 	fprintf (out, "@@ -%lu", orig_offset);
@@ -163,7 +169,8 @@ static unsigned long adjust_offsets_and_copy (long *offset, FILE *in,
 	fprintf (out, " +%lu", new_offset + *offset);
 	if (new_lines != 1)
 		fprintf (out, ",%lu", new_lines);
-	fprintf (out, " @@\n");
+	fprintf (out, " @@%s", trailing);
+	free (line);
 
 	/* Copy remaining lines of hunk */
 	count += copy_hunk (in, out, orig_lines, new_lines);
@@ -250,7 +257,7 @@ static long added_hunk (const char *meta, long offset, FILE *modify, FILE *t,
 	char *p = strchr (meta, '-'), *q;
 	unsigned long orig_offset = 0, new_offset;
 	unsigned long orig_count = 0, new_count = 0;
-	FILE *newhunk = tmpfile ();
+	FILE *newhunk = xtmpfile ();
 
 	if (!newhunk)
 		error (EXIT_FAILURE, errno, "Couldn't create temporary file");
@@ -331,6 +338,7 @@ static long removed_hunk (const char *meta, FILE *modify, FILE *t,
 	size_t linelen = 0;
 	unsigned long orig_offset, new_offset;
 	unsigned long orig_count, new_count;
+	int r;
 
 	*replaced = 0;
 	if (read_atatline (meta, &orig_offset, &orig_count,
@@ -343,6 +351,7 @@ static long removed_hunk (const char *meta, FILE *modify, FILE *t,
 	if (line[0] == '+' && line[1] == '@') {
 		/* Minimally correct modified @@ banners. */
 		unsigned long oo, no;
+		char *trailing;
 		if (read_atatline (line + 1, &oo, NULL, &no, NULL))
 			goto out;
 
@@ -353,6 +362,12 @@ static long removed_hunk (const char *meta, FILE *modify, FILE *t,
 			hunk->info->info_written = 1;
 		}
 
+		trailing = strchr (line + 1, '+');
+		trailing += strcspn (trailing, " \n");
+		if (*trailing == ' ')
+			trailing++;
+		trailing += strspn (trailing, "@");
+
 		if (oo != orig_offset)
 			no = new_offset + oo - orig_offset;
 		else
@@ -363,7 +378,7 @@ static long removed_hunk (const char *meta, FILE *modify, FILE *t,
 		fprintf (t, " +%lu", no);
 		if (new_count != 1)
 			fprintf (t, ",%lu", new_count);
-		fprintf (t, " @@\n");
+		fprintf (t, " @@%s", trailing);
 		goto out;
 	}
 
@@ -398,7 +413,8 @@ static long removed_hunk (const char *meta, FILE *modify, FILE *t,
 			if (!--morig_count)
 				break;
 
-			assert (getline (&line, &linelen, modify) != -1);
+			r = getline (&line, &linelen, modify);
+			assert (r != -1);
 		}
 
 		if (morig_count) {
@@ -415,7 +431,8 @@ static long removed_hunk (const char *meta, FILE *modify, FILE *t,
 				hunk->info->info_pending = 1;
 			*hunkp = hunk;
 
-			assert (getline (&line, &linelen, modify) != -1);
+			r = getline (&line, &linelen, modify);
+			assert (r != -1);
 			morig_count--;
 		}
 	}
@@ -448,10 +465,12 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 	unsigned long morig_offset, morig_count, mnew_offset, mnew_count;
 	char *line = NULL;
 	size_t linelen = 0;
-	FILE *t = tmpfile ();
+	FILE *t = xtmpfile ();
 	int t_written_to = 0;
 	unsigned long i, at = 1;
 	unsigned long replaced, unaltered;
+	char *trailing;
+	int r;
 
 	if (!t)
 		error (EXIT_FAILURE, errno, "Couldn't open temporary file");
@@ -459,26 +478,42 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 	rewind (modify);
 
 	fsetpos (original, &hunk->filepos);
-	assert (getline (&line, &linelen, original) != -1);
+	r = getline (&line, &linelen, original);
+	assert (r != -1);
 	if (hunk->info) {
-		assert (getline (&line, &linelen, original) != -1);
-		assert (getline (&line, &linelen, original) != -1);
+		r = getline (&line, &linelen, original);
+		assert (r != -1);
+		r = getline (&line, &linelen, original);
+		assert (r != -1);
 		at += 2;
 	}
-	assert (!read_atatline (line, &orig_offset, &orig_count,
-				&new_offset, &new_count));
+	r = read_atatline (line, &orig_offset, &orig_count,
+			   &new_offset, &new_count);
+	assert (!r);
 	calc_orig_count = orig_count;
 	calc_new_offset = new_offset;
 	calc_new_count = new_count;
 
-	assert (getline (&line, &linelen, modify) != -1);
-	assert (!read_atatline (line, &morig_offset, &morig_count,
-				&mnew_offset, &mnew_count));
+	trailing = strchr (line, '+');
+	trailing += strcspn (trailing, " \n");
+	if (*trailing == ' ')
+		trailing++;
+	trailing += strspn (trailing, "@");
+	trailing = xstrdup (trailing);
+
+	r = getline (&line, &linelen, modify);
+	assert (r != -1);
+	r = read_atatline (line, &morig_offset, &morig_count,
+			   &mnew_offset, &mnew_count);
+	assert (!r);
 	replaced = morig_count;
 
 	do {
 		/* Lines before the modification are unaltered. */
 		int trim = 0;
+		if (morig_offset < hunk->line_in_diff)
+			error (EXIT_FAILURE, errno, "Invalid changes made");
+
 		unaltered = morig_offset - hunk->line_in_diff;
 		if (!morig_count)
 			unaltered++;
@@ -492,7 +527,8 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 		fprintf (stderr, "First %lu lines unaltered\n", unaltered);
 #endif /* DEBUG */
 		for (i = unaltered; i; i--) {
-			assert (getline (&line, &linelen, original) != -1);
+			r = getline (&line, &linelen, original);
+			assert (r != -1);
 			fputs (line, t);
 			at++;
 			t_written_to = 1;
@@ -601,7 +637,8 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 #endif /* DEBUG */
 		while (replaced) {
 			replaced--;
-			assert (getline (&line, &linelen, original) != -1);
+			r = getline (&line, &linelen, original);
+			assert (r != -1);
 			at++;
 			switch (line[0]) {
 			case ' ':
@@ -662,8 +699,9 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 
 				while (replaced) {
 					replaced--;
-					assert (getline (&line, &linelen,
-							 original) != -1);
+					r = getline (&line, &linelen,
+						     original);
+					assert (r != -1);
 					at++;
 					switch (line[0]) {
 					case ' ':
@@ -683,8 +721,9 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 		if (getline (&line, &linelen, modify) == -1)
 			break;
 
-		assert (!read_atatline (line, &morig_offset, &morig_count,
-				        &mnew_offset, &mnew_count));
+		r = read_atatline (line, &morig_offset, &morig_count,
+				   &mnew_offset, &mnew_count);
+		assert (!r);
 		replaced = morig_count;
 	} while (!feof (modify));
 
@@ -724,16 +763,17 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 	}
 	rewind (t);
 
-	if (calc_orig_count && calc_new_count) {
+	if (calc_orig_count || calc_new_count) {
 		fprintf (out, "@@ -%lu", orig_offset);
 		if (calc_orig_count != 1)
 			fprintf (out, ",%lu", calc_orig_count);
 		fprintf (out, " +%lu", calc_new_offset + line_offset);
 		if (calc_new_count != 1)
 			fprintf (out, ",%lu", calc_new_count);
-		fprintf (out, " @@\n");
+		fprintf (out, " @@%s", trailing);
 	}
 
+	free (trailing);
 	while (getline (&line, &linelen, t) != -1)
 		fputs (line, out);
 
@@ -866,7 +906,7 @@ static int rediff (const char *original, const char *edited, FILE *out)
 	/* Run diff between original and edited. */
 	t = xpipe (DIFF, &child, "r", DIFF, "-U0",
 		   original, edited, NULL);
-	m = tmpfile ();
+	m = xtmpfile ();
 	if (m) {
 		size_t buffer_size = 10000;
 		char *buffer = xmalloc (buffer_size);
@@ -949,7 +989,7 @@ static int rediff (const char *original, const char *edited, FILE *out)
 			/* This meta hunk is the first pertaining to
 			 * the hunk in the original. */
 			first_hunk = pos;
-			t = tmpfile ();
+			t = xtmpfile ();
 		}
 
 		current_hunk = which;
@@ -1025,7 +1065,8 @@ int main (int argc, char *argv[])
 		syntax (1);
 
 	if (argc - optind == 1) {
-		char *p = xmalloc (PATH_MAX);
+		char *p = xmalloc (strlen (argv[0]) +
+				   strlen ("recountdiff") + 1);
 		char *f;
 		char **const new_argv = xmalloc (sizeof (char *) * argc);
 		memcpy (new_argv, argv, sizeof (char *) * argc);
@@ -1036,8 +1077,8 @@ int main (int argc, char *argv[])
 			f = p;
 		else f++;
 		strcpy (f, "recountdiff");
-		execv (new_argv[0], new_argv);
-		p = strdup (new_argv[0]);
+		execvp (new_argv[0], new_argv);
+		p = xstrdup (new_argv[0]);
 		f = strstr (p, "src/");
 		if (f) {
 			while (*(f + 4)) {
